import abc
from datetime import datetime
from .component import BpmnComponent
import uuid
import logging

logger = logging.getLogger(__name__)

class Task(BpmnComponent):

    def __init__(self,activity_data,process_instance):
        self.id = activity_data.get("@id")
        self.activity_data = activity_data
        self.process_instance = process_instance
        self.name = self.activity_data.get("@name")
        self.activity_id = str(uuid.uuid4())

    @abc.abstractmethod
    def execute(self,context,payload):
        pass
    
    def _execute(self,context,payload):
        self.context = context
        self.payload = payload
        name = self.name
        if ( self.context.get(name) == None ) :
            self.context[name] = {}

        if ( self.payload.get(name) == None ) :
            self.payload[name] = {}

        self.task_context = self.context[name]
        self.task_context["start_time"] = datetime.now()
        self.task_context["name"] = name
        self.task_context["id"] = self.id
        self.task_context["activity_id"] = self.activity_id
        self.payload_task = payload[name]
        
        logger.info({
            "message": f"Executing service task:{name}",
            "data" : self.activity_data
        })

    def outgoing_flow_success(self,flow_id):
        context = self.context
        if context.get(flow_id) == None:
            context[flow_id] = {}

        context[flow_id]["start_time"] = datetime.now()        
        context[flow_id]["end_time"] = datetime.now()
        context[flow_id]["id"] = flow_id


    def get_outgoing_activities(self):
        
        outgoingflowid = self.activity_data.get("bpmn:outgoing")

        target_activity_id = None
        target_activity_type = None
        target_activity_data = None
        for seq_flow in self.process_instance.process_definition.get("bpmn:sequenceFlow"):
            if(seq_flow.get("@id") == outgoingflowid):
                self.outgoing_flow_success(outgoingflowid)
                target_activity_id = seq_flow.get("@targetRef")

        for component_name in self.process_instance.process_definition:
            
            if type(self.process_instance.process_definition[component_name]) == type(dict()):
                component = self.process_instance.process_definition[component_name]
                if component.get("@id") == target_activity_id:
                    target_activity_data = component
                    target_activity_type = component_name
            
            if type(self.process_instance.process_definition[component_name]) == type([]):
                for component in self.process_instance.process_definition[component_name]:
                    if component.get("@id") == target_activity_id:
                        target_activity_data = component
                        target_activity_type = component_name

        return [(target_activity_type,target_activity_id,target_activity_data)]
        


import logging
from datetime import datetime
from .task import Task

logger = logging.getLogger(__name__)

class ServiceTask(Task):
    def __init__(self,activity_data,process_instance):
        
        super().__init__(activity_data,process_instance)

        self.activity_data = activity_data
        self.process_instance = process_instance

    def execute(self,context,payload):
        self._execute(context,payload)
        name = self.name
        task = self.task_context
        task["status"] = "STARTED"
        payload_task = self.payload_task

        if self.process_instance.handler != None:
            if hasattr(self.process_instance.handler,f"on_enter_task"):
                getattr(self.process_instance.handler,f"on_enter_task")(context = context,task = payload_task, payload = payload, process_instance = self.process_instance)
                
            if hasattr(self.process_instance.handler,f"on_enter_{name}"):
                getattr(self.process_instance.handler,f"on_enter_{name}")(context = context,task = payload_task, payload = payload, process_instance = self.process_instance)
            
            if hasattr(self.process_instance.handler,f"on_{name}"):
                getattr(self.process_instance.handler,f"on_{name}")(context = context,task = payload_task, payload = payload, process_instance = self.process_instance)

            if hasattr(self.process_instance.handler,f"on_exit_{name}"):
                getattr(self.process_instance.handler,f"on_exit_{name}")(context = context,task = payload_task, payload = payload, process_instance = self.process_instance)

            if hasattr(self.process_instance.handler,f"on_exit_task"):
                getattr(self.process_instance.handler,f"on_exit_task")(context = context,task = payload_task, payload = payload, process_instance = self.process_instance)
        
        context[name]["end_time"] = datetime.now()
        task["status"] = "COMPLETED"
        
        self.process_instance.evaluate_results(self.get_outgoing_activities())


import logging

from datetime import datetime
from .task import Task

logger = logging.getLogger(__name__)

class EndEvent(Task):

    def execute(self,context,payload):

        self._execute(context,payload)

        self.task_context["start_time"] = datetime.now()
        self.task_context["end_time"] = datetime.now()

        context["status"] = "COMPLETED"
        
        self.process_instance.evaluate_results(self.get_outgoing_activities())


import abc
from .gateway import Gateway
from datetime import datetime

import logging

logger = logging.getLogger(__name__)

class ExclusiveGateway(Gateway):
    
    def execute(self,context,payload):
        self._execute(context,payload)
        name = self.name
        task = self.task_context
        
        context[name]["start_time"] = datetime.now()        
        context[name]["end_time"] = datetime.now()
        self.process_instance.evaluate_results(self.get_outgoing_activities())

    def outgoing_flow_success(self,flow_id):
        context = self.context
        if context.get(flow_id) == None:
            context[flow_id] = {}

        context[flow_id]["start_time"] = datetime.now()        
        context[flow_id]["end_time"] = datetime.now()
        context[flow_id]["id"] = flow_id

    def is_atleast_one_incoming_complete(self):
        incomingflowids = self.activity_data.get("bpmn:incoming")
        source_activity_refs=[]
        if type(incomingflowids) == type([]):
            for incomingflowid in incomingflowids:
                for seq_flow in self.process_instance.process_definition.get("bpmn:sequenceFlow"):
                    if(seq_flow.get("@id") == incomingflowid):
                        source_activity_refs.append(seq_flow.get("@sourceRef"))

        source_activity_names = []
        for source_activity_ref in source_activity_refs:
            for component_name in self.process_instance.process_definition:
                if type(self.process_instance.process_definition[component_name]) == type(dict()):
                    component = self.process_instance.process_definition[component_name]
                    if component.get("@id") == source_activity_ref:
                        source_activity_names.append(component.get("@name"))

                if type(self.process_instance.process_definition[component_name]) == type([]):
                    for component in self.process_instance.process_definition[component_name]:
                        if component.get("@id") == source_activity_ref:
                            source_activity_names.append(component.get("@name"))

        for source_activity_name in source_activity_names:
            if(self.context.get(source_activity_name,{}).get("status") == "COMPLETED"):
                return True

        return False

    def is_all_incoming_complete(self):
        incomingflowids = self.activity_data.get("bpmn:incoming")
        source_activity_refs=[]
        if type(incomingflowids) == type([]):
            for incomingflowid in incomingflowids:
                for seq_flow in self.process_instance.process_definition.get("bpmn:sequenceFlow"):
                    if(seq_flow.get("@id") == incomingflowid):
                        source_activity_refs.append(seq_flow.get("@sourceRef"))

        source_activity_names = []
        for source_activity_ref in source_activity_refs:
            for component_name in self.process_instance.process_definition:
                if type(self.process_instance.process_definition[component_name]) == type(dict()):
                    component = self.process_instance.process_definition[component_name]
                    if component.get("@id") == source_activity_ref:
                        source_activity_names.append(component.get("@name"))

                if type(self.process_instance.process_definition[component_name]) == type([]):
                    for component in self.process_instance.process_definition[component_name]:
                        if component.get("@id") == source_activity_ref:
                            source_activity_names.append(component.get("@name"))

        for source_activity_name in source_activity_names:
            if(self.context.get(source_activity_name,{}).get("status","NOT_DEFINED") != "COMPLETED"):
                return False

        return True

    def get_outgoing_activities(self):
        context = self.context
        payload = self.payload
        outgoingflowids = self.activity_data.get("bpmn:outgoing")
        
        if type(outgoingflowids) != type([]):
            outgoingflowids = [outgoingflowids]

        target_activity_ids=[]
        if type(outgoingflowids) == type([]):
            for outgoingflowid in outgoingflowids:
                for seq_flow in self.process_instance.process_definition.get("bpmn:sequenceFlow"):
                    if(seq_flow.get("@id") == outgoingflowid):
                        if seq_flow.get("bpmn:conditionExpression",{}).get("#text") != None:
                            if eval(seq_flow.get("bpmn:conditionExpression",{}).get("#text")) == True:
                                target_activity_ids.append(seq_flow.get("@targetRef"))
                                self.outgoing_flow_success(outgoingflowid)

                                # break at the first successful condition
                                break
                        else:
                            # If there is no condition, choose the first path
                            target_activity_ids.append(seq_flow.get("@targetRef"))
                            self.outgoing_flow_success(outgoingflowid)
                            break
                        

        targets = []

        logger.info(target_activity_ids)

        for target_activity_id in target_activity_ids:
            for component_name in self.process_instance.process_definition:
                if type(self.process_instance.process_definition[component_name]) == type(dict()):
                    component = self.process_instance.process_definition[component_name]
                    if component.get("@id") == target_activity_id:
                        target_activity_data = component
                        target_activity_type = component_name

                        targets.append((target_activity_type,target_activity_id,target_activity_data))

                if type(self.process_instance.process_definition[component_name]) == type([]):
                    for component in self.process_instance.process_definition[component_name]:
                        if component.get("@id") == target_activity_id:
                            target_activity_data = component
                            target_activity_type = component_name
                            targets.append((target_activity_type,target_activity_id,target_activity_data))


        return targets